///////////////////////////////////////////////////////////////////////////////
//	Module			: CEif.C
//	Description		: Implementation of CEif class. 
//                    This class is a delived from CSerial class and
//                    provides functionality to communicate with EIF.
//
//	Compiler		: gcc 4.4.7++
//	Author			: T.Matsuzawa of JEI Matsumoto Lab.
//	E-mail			: tmatsu.jei@nifty.com
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Maintenance:
//	Jan.15, 2014	Renewal for CentOS6.5/64bit
///////////////////////////////////////////////////////////////////////////////	
#include 	<iostream>
#include 	<iomanip>
#include 	<sstream>
using namespace std;

#include    <sys/ioctl.h>
#include    <pthread.h>
#include 	<unistd.h>
#include	<stdio.h>
#include	<stdlib.h>

#include 	"../../include/pifg/libPIF.h"
//#include 	"pifg/common.h"
//#include 	"pif/CEif.h"
//#include "pifg/CLog.h"

int timeOutRetry = 0;

//	External classes ________________________________________________________
extern CLog	oLog;

#define		DBG			1		// 1 : debug message is available

//	Extarnal functions ______________________________________________________

int	getTime();					// get current time in sec.
///extern  pthread_mutex_t     mutex_eif;

void logToGenesis(char*);
//  Local variables _________________________________________________________

static CONF	config[] = {
	{ 1,  5, defImageDensity, 	SR_ID,  1},	// #00: Image Density
	{30,144, defFormHeight, 	SR_FH,  1},	// #01: Form Height
	{60,180, defFormWidth, 		SR_FW,  1},	// #02: Form Width
	{ 1,  8, defSuctionDuty, 	SR_SFD, 1},	// #03: Suction Fan Duty
	{ 0,  3, defTransferCurrent,SR_TRC, 1},	// #04: Transfer Current
	{ 0,  1, defSeparatorCHCntl,SR_SCC, 1},	// #05: Separator CH control
	{ 0,  4, defAIDCMode, 		SR_AIM, 0},	// #06: AIDC Mode
	{ 0,  5, defResolution, 	SR_RES, 1},	// #07: Resolution
	{ 0,254, defFineAdjust, 	SR_VFA, 1},	// #08: Vertical Fine Adjustment
	{ 0,254, defFineAdjust, 	SR_HFA, 1},	// #09: Horizontal Fine Adjustment
	{ 0, 66, defMargin, 		SR_TM, 1},	// #10: Top Margin
	{ 0, 66, defMargin, 		SR_BM, 1},	// #11: Bottom Margin
	{ 0,100, defMargin, 		SR_LM, 1},	// #12: Left Margin
	{ 0,  1, 0, 				SR_RVM, 1},	// #13: Reverse Mode
	{ 1,  4, defEnergy4Fusion,	SR_EFF, 1},	// #14: Energy For Fusion
	{-1,-1, -1, -1},						// #xx: end mark
};

static PrinterConfNdx prnConfNdx[] = {
	{SR_ID, CNF_ID},	// #00: Image Density
	{SR_FH, CNF_FH},	// #01: Form Height
	{SR_FW,	CNF_FW},	// #02: Form Width
	{SR_SFD,CNF_SFD},	// #03: Suction Fan Duty
	{SR_TRC,CNF_TRC},	// #04: Transfer Current
	{SR_SCC,CNF_SCC},	// #05: Separator CH control
	{SR_AIM,CNF_AIM},	// #06: AIDC Mode
	{SR_RES,CNF_RES},	// #07: Resolution
	{S_VOR,	CNF_VOR},	// #08: Vertical Origin
	{S_HOR,	CNF_HOR},	// #09: Horizontal Origin
	{SR_VFA,CNF_VFA},	// #10: Vertical Fine Adjustment
	{SR_HFA,CNF_HFA},	// #11: Horizontal Fine Adjustment
	{SR_TM,	CNF_TM},	// #12: Top Margin
	{SR_BM,	CNF_BM},	// #13: Bottom Margin
	{SR_LM,	CNF_LM},	// #14: Left Margin
	{SR_RVM,CNF_RVM},	// #15: Reverse Mode
	{SR_EFF,CNF_EFF},	// #16: Energy For Fusion
	{R_VOR,	CNF_VOR},	// #08: Vertical Origin
	{R_HOR,	CNF_HOR},	// #09: Horizontal Origin
	{-1,-1		}	// #xx: end mark
};

//const int offset = 0x8000;	// found out that this value prevents debugging as well as 0x4000, 0x2000
const int offset = 0x1000;		// this is the value that is friendlier for the debugging to work.

void logToGenesis(char* message)
{
	FILE* genesisLog;
	time_t rawtime;
	struct tm* timeinfo;
	char buffer[80];

	time(&rawtime);
	timeinfo = localtime(&rawtime);

	genesisLog = fopen("/home/Genesis/error.log","at");
	if(genesisLog) {
		strftime(buffer,80,"%02d/%02m/%02Y %02I:%02M:%02S ", timeinfo);
		string logMessage(buffer);
		logMessage.append(message);
		fclose(genesisLog);
	}

	genesisLog = 0;
}

//	Global methods __________________________________________________________
/*
	<method>		CEif
	<description>	Constructor 
	<args>			int fd = file descriptor
*/
CEif::CEif(int fd)
{
	LOG("libPIF",LogInfo,"Creating CEif class instance ...");
    nFd = fd; 						// save a file descriptor
	port = 0;						// use channel-A
	nActive = 0;					// EIF inactive

//	initialize mutes
///	 pthread_mutex_init(&mutex_eif, NULL);

//	set default of configurations
	setDefaultConfig();

#if	DBG
	nDBG = 1;						// enable debug messages
#else
	nDBG = 0;						// disable debug messages
#endif
//	findConf(1234);
//	findPrinterConfNdx(SR_ID);
//	findPrinterConfNdx(S_VOR);
//	findPrinterConfNdx(R_VOR);
//	findPrinterConfNdx(3355);
//	exit(0);
}

int CEif::findConf(int id)
{
	int sizeConfigAll,sizeConfigRow,rows, confNdx;
	
	sizeConfigAll = sizeof(config);
	sizeConfigRow = sizeof(CONF);
	rows = sizeConfigAll/sizeConfigRow;
	//printf("confall = %d, confrow = %d, confall/confrow = %d\n",
		//   sizeConfigAll,sizeConfigRow,rows);
	for(confNdx = 0; confNdx < rows; confNdx++)
	{
		if(config[confNdx].cmd == id)
		{
			break;
		}
	}
	if(confNdx > (rows-1)) confNdx = rows - 1;
///	printf("id = %d, confNdx = %d\n",id, confNdx);
	
	return confNdx;
}

int CEif::getnActive()
{
	return nActive;
}

int CEif::findPrinterConfNdx(int cmd)
{
	int sizeConfigAll,sizeConfigRow,rows, n, confNdx;
	
	sizeConfigAll = sizeof(prnConfNdx);
	sizeConfigRow = sizeof(PrinterConfNdx);
	rows = sizeConfigAll/sizeConfigRow;
	//printf("confall = %d, confrow = %d, confall/confrow = %d\n",
		//   sizeConfigAll,sizeConfigRow,rows);
	for(n = 0; n < rows; n++)
	{
		if(prnConfNdx[n].cmdId == cmd)
		{
			break;
		}
	}
	if(n > (rows-1)) n = rows - 1;
	confNdx = prnConfNdx[n].confNdx;
///	printf("id = %d, n = %d, confNdx = %d\n",cmd, n, confNdx);
	
	return confNdx;
}

/*
	<method>		getCTS	
	<description> 	get CTS signal	
					Returns CST signal status in the LSR
	<args>			none
	<return>		int = true(1) : active 
						  false(0): inactive 
*/
int	
CEif::getCTS()
{
	int		i, cts1, cts2;
	
	int 	ret = false;
	cts1 =  getRegister(R_UART_MSRA) & 0x10;		// read Modem Status reg
	for(i = 0; i < 10; i++){
		usleep(0);									// sleep a little
		cts2 =  getRegister(R_UART_MSRA) & 0x10;	// read Modem Status reg
		if(cts1 == cts2){
			if(cts1)
				ret = true;
			break;
		}
		cts1 = cts2;
	}
	return(ret);
}

/*
	<method>		getConfig	
	<description>	receive configuration to prionter  
	<args>			none
	<return>		none
*/
void 
CEif::getConfig()
{
	int		i = 0;
	CONF*	ptr = config;

//	receive printer configurations from printer
	while(ptr[i].min >= 0){
		putRequest(ptr[i].cmd,i,0);	
		i++;
	}
}

/*
	<method>		getPageEnd
	<description>	Get number of PAGEEND status
					Returns number of PAGEND status
	<args>			none
	<return>		int = number of PAGEND status 
*/
int	
CEif::getPageEnd()
{
	return(ioctl(nFd, PIF_RD_PAGEND));
}

/*
	<method>		releaseWait
	<description>	Release any wait state (mainly page end)
	<args>			none
	<return>		none
*/
void
CEif::releaseWait()
{
	ioctl(nFd, PIF_WAIT_RELEASE);
}

/*
	<method>		getTotalPageEnd
	<description>	Get total number of PAGEEND status
					Returns total number of PAGEND status
	<args>			none
	<return>		int = number of PAGEND status 
*/
int	
CEif::getTotalPageEnd()
{
	return(ioctl(nFd, PIF_RD_INTC, 11));
}

int
CEif::getWarning()
{
	int	val;

	val = (putRequest(R_PCS,-1,0) & 0x3f) << 9;
	val |= putRequest(R_WRN,-1,0) & 0x1ff;
	return val;
}

/*
	<method>		putOperation
	<description>	Sends a operation command to EIF
	<args>			int	cmd = command
					int par = parameter	
	<return>		int = 	EIF_OK(0)
							EIF_TX_ERROR(-1)		
							EIF_RX_TIMEOUT(-2)
*/
int	
CEif::putOperation(int cmd, int par, char *s)
{
	char	buf[8];
	int		ret;
	ostringstream os;
	char	dst[256];
	int retry;

	concats(dst,s,"ptOp");
	if(!par)								// OFF command
		sprintf(buf,"\002F%02d\003",cmd);	
	else
		sprintf(buf,"\002N%02d\003",cmd);	// ON command
	for(retry = 0; retry < EIF_MAX_RETRY; retry++){
		if((ret = putCommand(buf, EIF_TIMEOUT1,dst)) >= 0 )
			break;
		else if(nActive == false)
			break;
		os.str("");
	//	os << "EIF Operation command Retry = " << retry;
		os << "EIF Operation cmd(" << dec << cmd << ") Retry = " << dec << retry << "(" << ret << ")";
		LOG("libPIF",LogDebug,os.str());
	}
//	if(retry >= EIF_MAX_RETRY)
//			logToGenesis("EIF Rx Time Out !");
	return(ret);
}

/*
	<method>		putSetup
	<description>	Tx setup command to EIF
	<args>			int	cmd = command
					int par = parameter	
					int cnf = configuration #
	<return>		int EIF_OK(0)
						EIF_TX_ERROR(-1)		
						EIF_RX_TIMEOUT(-2)
*/
int	
CEif::putSetup(int cmd, int par, int cnf, char *s)
{
	int		ret;
	char	buf[16];
	ostringstream	os;
	char	dst[256];
	int retry;
	
	concats(dst,s,"ptSU1");
	sprintf(buf,"\002S%02d%X\003",cmd,par);	
	for(retry = 0; retry < EIF_MAX_RETRY; retry++){
		if((ret = putCommand(buf, EIF_TIMEOUT2,dst)) >= 0 )
			break;
		else if(nActive == false)
			break;
		os.str("");
	//	os << "EIF Setup command Retry = " << retry;
		os << "EIF Setup1 cmd(" << dec << cmd << ") Retry = " << dec << retry << "(" << ret << ")";
		LOG("libPIF",LogDebug,os.str());
	}
	if(cnf >= 0)
		nPrinterConf[cnf] = par;
	return(ret);
}

int	
CEif::putSetup(int cmd, int par, char *s)
{
	int		ret;
	char	buf[16];
	ostringstream	os;
	int cnfNdx;
	char	dst[256];
	int retry;
	
	concats(dst,s,"ptSU2");
	cnfNdx = findPrinterConfNdx(cmd);
	ret = 0;
	sprintf(buf,"\002S%02d%X\003",cmd,par);	
	for(retry = 0; retry < EIF_MAX_RETRY; retry++){
		if((ret = putCommand(buf, EIF_TIMEOUT2,dst)) >= 0 )
			break;
		else if(nActive == false)
			break;
		os.str("");
	//	os << "EIF Setup command Retry = " << retry;
		os << "EIF Setup2 cmd(" << dec << cmd << ") Retry = " << dec << retry << "(" << ret << ")";
		LOG("libPIF",LogDebug,os.str());
	}
	if(cnfNdx >= 0)
		nPrinterConf[cnfNdx] = par;
	///return(ret);
	return(ret >= 0);
}


/*
	<method>		putDMC
	<description>	Tx setup command to EIF (Direct MC)
	<args>			int	cmd = command
					int par = parameter	
	<return>		int EIF_OK(0)
						EIF_TX_ERROR(-1)		
						EIF_RX_TIMEOUT(-2)
*/
int	
CEif::putDMC(int cmd, int par, char *s)
{
	int		ret;
	char	buf[16];
	ostringstream	os;
	char	dst[256];
	int retry;
	
	concats(dst,s,"ptMC");
	sprintf(buf,"\002S%02d%02d%X\003",S_DMC,cmd,par);	
	for(retry = 0; retry < EIF_MAX_RETRY; retry++){
		if((ret = putCommand(buf, EIF_TIMEOUT2,dst)) >= 0 )
			break;
		else if(nActive == false)
			break;
		os.str("");
		os << "EIF Setup DMC command Retry = " << retry;
		LOG("libPIF",LogDebug,os.str());
	}
//	if(retry >= EIF_MAX_RETRY)
//			logToGenesis("EIF Rx Time Out !");
	return(ret);
}

/*
	<method>		putRequest
	<description>	Tx status request command to EIF
	<args>			int	cmd = command
					int cnf = configuration #
					int flag = 0 : RX from printer
							   1 : return value in the work area
	<return>		int = RX value (flag = 0)
						  value in the work area (flag = 1)	
*/
int	
CEif::putRequest(int cmd, int cnf, int flag, char *s)
{
	int		ret;
	char	buf[8];
	ostringstream	os;
	char	dst[256];
	int retry;

	concats(dst,s,"ptRq1");
	if(!flag){
		sprintf(buf,"\002R%02d\003",cmd);	
		for(retry = 0; retry < EIF_MAX_RETRY; retry++){
			if((ret = putCommand(buf, EIF_TIMEOUT1,dst)) >= 0 )
				break;
			else if(nActive == false)
			{
				nActive = true;
				timeOutRetry = 0;
				break;
			}
			os.str("");
		//	os << "EIF Request cmd(" << hex << cmd << ") Retry = " << dec << retry << "(" << ret << ")";
			os << "EIF Request cmd(" << dec << cmd << ") Retry = " << dec << retry << "(" << ret << ")";
			LOG("libPIF",LogDebug,os.str());
		}
		if((cnf >= 0) && (ret >= 0))
			nPrinterConf[cnf] = ret;
	}
	else{
		ret = nPrinterConf[cnf];
	}
	return(ret);
}

int	
CEif::putRequest(int cmd, int flag, char *s)
{
	int		ret;
	char	buf[8];
	ostringstream	os;
	int cnfNdx;
	char	dst[256];
	int retry = 0;

	concats(dst,s,"ptRq2");
	cnfNdx = findPrinterConfNdx(cmd);
	if(!flag){
		sprintf(buf,"\002R%02d\003",cmd);	
		for(retry = 0; retry < EIF_MAX_RETRY; retry++){
			if((ret = putCommand(buf, EIF_TIMEOUT1,dst)) >= 0 )
				break;
			else if(nActive == false)
			{
				nActive = true;
				timeOutRetry = 0;
				break;
			}
			os.str("");
		//	os << "EIF Request cmd(" << hex << cmd << ") Retry = " << dec << retry << "(" << ret << ")";
			os << "EIF Request cmd(" << dec << cmd << ") Retry = " << dec << retry << "(" << ret << ")";
			LOG("libPIF",LogDebug,os.str());
		}
		if((cnfNdx >= 0) && (ret >= 0))
			///nPrinterConf[cnf] = ret;
			nPrinterConf[cnfNdx] = ret;
	}
	else
	{
		ret = nPrinterConf[cnfNdx];
	}
	
	if (cmd == 83)
	{
		os.str("");
		os << "EIF Request cmd(" << dec << cmd << ") Return = (" << dec << ret << ")";
		LOG("libPIF",LogDebug,os.str());
	}
	return(ret);
}

/*
	<method>		resetPagheEnd
	<description>	Reset PAGEEND interrupt counter
	<args>			none
	<return>		none
*/
void 
CEif::resetPageEnd()
{
	ioctl(nFd, PIF_RST_PAGEND);
}

/*
	<method>		setConfig	
	<description>	transmit configuration to prionter  
	<args>			none
	<return>		none
*/
void 
CEif::setConfig()
{
	int		i = 0;
	CONF*	ptr = config;

//	set dafault for printer configurations
	while(ptr[i].min >= 0){
		if(ptr[i].flag)
			putSetup(ptr[i].cmd, nPrinterConf[i],-1);	
		i++;
	}
}

//	Private mathods _________________________________________________________
/*
	<method>		putCommand
	<description>	sends a command to EIF
	<args>			char* buf = pointer to buffer
					int timeout = time out (sec.)
	<return>		int =	EIF_OK(0)
							EIF_TX_ERROR(-1)		
							EIF_RX_TIMEOUT(-2)
*/
int	
CEif::putCommand(char* buf, int timeout, char *s)
{
	int		ret = true;
	char	rxbuf[16],*ptr;
	int		start;
	int		c,n;
	char	cmd;
	char	dst[256];
	char	str[256];
	int		lid;
	ostringstream	retryLog;
	ostringstream	os;

	if(nActive == false)
		return(EIF_RX_TIMEOUT);

//	Tx command
	concats(dst,s,"ptCmd");
	lid=BeginCmdSeq(dst);
	sprintf(str,"ptCmd[%d]",lid);
	concats(dst,s,str);
	if(nActive){
//		cout << "TX: " << buf << endl << flush;
		//if((ret = puts(buf,dst)) != true){
		if((ret = puts(buf,offset,dst)) != true){	
			LOG("libPIF",LogErr,"EIF Tx Error !");
///			pthread_mutex_unlock(&mutex_eif);
			EndCmdSeq(lid, dst);
			return(EIF_TX_ERROR);					// return if failed
		}
//	Wait for answer
		start = getTime();							// start timer
		while(true){
			if (timeOutRetry > 0)
				start = getTime();
			if((c = getchar(offset,dst)) == C_STX){ //start of text
				ptr = rxbuf;
				cmd = getchar(offset,dst);					// RX command 
				for(n = 0; n < 15; n++){			// RX parameter
					if((c = getchar(offset,dst)) == C_ETX){ //end of text
						*ptr = '\0';
						break;
					}
					*ptr++ = toupper(c);
				}	
				break;
			}
			else if(abs(getTime() - start) > timeout)
			{
				timeOutRetry++;
				retryLog.str("");
				retryLog << "EIF time out Retry = (" << dec << timeOutRetry << ") Return = (" << dec << ret << ")";
				LOG("libPIF",LogDebug,retryLog.str());
				
				if (timeOutRetry > EIF_MAX_RETRY)
				{
					LOG("libPIF",LogErr,"EIF Rx Time Out !");
					EndCmdSeq(lid, dst);
					nActive = false;
					return(EIF_RX_TIMEOUT);				// return if error
				}
			/*	else
				{
					return(EIF_RX_TIMEOUT);
				}*/
			}
		}

//	Analyze RX packet
		if(cmd == ST_ANSWER)						// RX answer
			ret = rxbuf[0] - '0';	
		else if(cmd == ST_STATUS){					// RX status
			for(ret = 0,c = 0; c < n; c++) {
				ret <<= 4;							// *= 16
				if(rxbuf[c] > '9')
					ret += (rxbuf[c] - '7') ;
				else
					ret += (rxbuf[c] - '0');
			}
		}
		else										// illegal status
			ret = EIF_RX_ILLEGAL;
	}
	else
		ret = EIF_RX_TIMEOUT;
	
	if (ret < -3)
	{
		os << "rxStatus:" << rxbuf << "(" << n << ")";
		LOG("libPIF",LogInfo,os.str().c_str());
	}
//cout << "\tRX: " << ret << endl << flush;
	EndCmdSeq(lid, dst);
	return (ret);
}

/*
	<method>		setDefaultConfig
	<description>	Set default value of configurations
	<args>			none
	<return>		none
*/
void 
CEif::setDefaultConfig()
{
	int		i = 0;
	CONF*	ptr = config;

//	set dafault for printer configurations

	while(ptr[i].min >= 0){
		nPrinterConf[i] = ptr[i].def;		// set default	
		i++;
	}
}

